import logging
import sys
from functools import wraps
from typing import Dict, List, Tuple

import numpy as np
from scipy.optimize import linear_sum_assignment

from nucleus.annotation import BoxAnnotation, PolygonAnnotation

from .custom_types import BoxOrPolygonAnnotation, BoxOrPolygonPrediction

try:
    from shapely.geometry import Polygon
except (ModuleNotFoundError, OSError):
    from ..package_not_installed import PackageNotInstalled

    Polygon = PackageNotInstalled


from .base import ScalarResult
from .errors import PolygonAnnotationTypeError


def polygon_annotation_to_shape(
    annotation: BoxOrPolygonAnnotation,
) -> Polygon:
    if isinstance(annotation, BoxAnnotation):
        xmin = annotation.x - annotation.width / 2
        xmax = annotation.x + annotation.width / 2
        ymin = annotation.y - annotation.height / 2
        ymax = annotation.y + annotation.height / 2
        return Polygon(
            [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
        )
    elif isinstance(annotation, PolygonAnnotation):
        return Polygon([(point.x, point.y) for point in annotation.vertices])
    else:
        raise PolygonAnnotationTypeError()


def _iou(annotation: Polygon, prediction: Polygon) -> float:
    intersection = annotation.intersection(prediction).area
    union = annotation.area + prediction.area - intersection
    return intersection / max(union, sys.float_info.epsilon)


def _iou_matrix(
    annotations: List[Polygon], predictions: List[Polygon]
) -> np.ndarray:
    iou_matrix = np.empty((len(predictions), len(annotations)))
    for i, prediction in enumerate(predictions):
        for j, annotation in enumerate(annotations):
            iou_matrix[i, j] = _iou(annotation, prediction)
    return iou_matrix


def _iou_assignments_for_same_reference_id(
    annotations: List[BoxOrPolygonAnnotation],
    predictions: List[BoxOrPolygonPrediction],
    iou_threshold: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    # Matches annotations and precitions of the same reference ID.
    # Returns a tuple of the list of all IoU values of valid assignments, a
    # list of the indices of predictions matched to annotations (-1 if
    # unmatched), and a list of all indices of annotations matched to
    # predictions.

    # Check that all annotations and predictions have same reference ID.
    reference_ids = set(annotation.reference_id for annotation in annotations)
    reference_ids |= set(prediction.reference_id for prediction in predictions)
    assert (
        len(reference_ids) <= 1
    ), "Expected annotations and predictions to have same reference ID."

    # Convert annotation and predictions to shapely.geometry.Polygon objects
    polygon_annotations = list(map(polygon_annotation_to_shape, annotations))
    polygon_predictions = list(map(polygon_annotation_to_shape, predictions))

    invalid_anns = [
        ann
        for ann, poly in zip(annotations, polygon_annotations)
        if not poly.is_valid
    ]
    invalid_preds = [
        pred
        for pred, poly in zip(predictions, polygon_predictions)
        if not poly.is_valid
    ]

    if invalid_anns or invalid_preds:
        # Filter out invalid polys
        polygon_annotations = [
            poly
            for ann, poly in zip(annotations, polygon_annotations)
            if poly.is_valid
        ]
        polygon_predictions = [
            poly
            for pred, poly in zip(predictions, polygon_predictions)
            if poly.is_valid
        ]
        invalid_dataset_ids = set(
            ann.reference_id for ann in invalid_anns
        ).union(set(pred.reference_id for pred in invalid_preds))
        # TODO(gunnar): change to .id once field is surfaced)
        logging.warning(
            "Invalid polygons for dataset items: %s Annotations:%s, predictions: %s",
            invalid_dataset_ids,
            [a.annotation_id for a in invalid_anns],
            [p.annotation_id for p in invalid_preds],
        )

    # Compute IoU matrix and set IoU values below the threshold to 0.
    iou_matrix = _iou_matrix(polygon_annotations, polygon_predictions)
    iou_matrix[iou_matrix < iou_threshold] = 0

    # Match annotations and predictions using linear sum assignment and filter out
    # values below the threshold.
    matched_0, matched_1 = linear_sum_assignment(-iou_matrix)
    iou_assigns = iou_matrix[matched_0, matched_1]
    valid_idxes = iou_assigns >= iou_threshold
    iou_assigns = iou_assigns[valid_idxes]

    matched_0 = matched_0[valid_idxes]
    matched_1 = matched_1[valid_idxes]
    anno_to_pred = -np.ones(len(annotations))
    pred_to_anno = -np.ones(len(predictions))
    anno_to_pred[matched_1] = matched_0
    pred_to_anno[matched_0] = matched_1

    return iou_assigns, anno_to_pred, pred_to_anno


def group_boxes_or_polygons_by_reference_id(
    annotations: List[BoxOrPolygonAnnotation],
    predictions: List[BoxOrPolygonPrediction],
) -> Dict[
    str, Tuple[List[BoxOrPolygonAnnotation], List[BoxOrPolygonPrediction]]
]:
    """Groups input annotations and predictions by reference_id.

    Args:
        annotations: list of input annotations
        predictions: list of input predictions

    Returns:
        Mapping from each reference_id to (annotations, predictions) tuple.
    """
    reference_ids = set(annotation.reference_id for annotation in annotations)
    reference_ids |= set(prediction.reference_id for prediction in predictions)
    grouped: Dict[
        str, Tuple[List[BoxOrPolygonAnnotation], List[BoxOrPolygonPrediction]]
    ] = {reference_id: ([], []) for reference_id in reference_ids}
    for annotation in annotations:
        grouped[annotation.reference_id][0].append(annotation)
    for prediction in predictions:
        grouped[prediction.reference_id][1].append(prediction)
    return grouped


def group_boxes_or_polygons_by_label(
    annotations: List[BoxOrPolygonAnnotation],
    predictions: List[BoxOrPolygonPrediction],
) -> Dict[
    str, Tuple[List[BoxOrPolygonAnnotation], List[BoxOrPolygonPrediction]]
]:
    """Groups input annotations and predictions by label.

    Args:
        annotations: list of input box or polygon annotations
        predictions: list of input box or polygon predictions

    Returns:
        Mapping from each label to (annotations, predictions) tuple
    """
    labels = set(annotation.label for annotation in annotations)
    labels |= set(prediction.label for prediction in predictions)
    grouped: Dict[
        str, Tuple[List[BoxOrPolygonAnnotation], List[BoxOrPolygonPrediction]]
    ] = {label: ([], []) for label in labels}
    for annotation in annotations:
        grouped[annotation.label][0].append(annotation)
    for prediction in predictions:
        grouped[prediction.label][1].append(prediction)
    return grouped


def iou_assignments(
    annotations: List[BoxOrPolygonAnnotation],
    predictions: List[BoxOrPolygonPrediction],
    iou_threshold: float,
) -> np.ndarray:
    """Matches annotations and predictions based on linear sum cost and returns the
    intersection-over-union values of the matched annotation-prediction pairs, subject
    to the specified IoU threshold. Note that annotations and predictions from
    different reference_ids will not be matched with one another.
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linear_sum_assignment.html

    Args:
        annotations: list of box or polygon annotations
        predictions: list of box or polygon predictions
        iou_threshold: the intersection-over-union threshold for an
            annotation-prediction pair to be considered a match.

    Returns:
        1D numpy array that contains the IoU values of the matched pairs.
    """
    grouped_inputs = group_boxes_or_polygons_by_reference_id(
        annotations, predictions
    )
    iou_assigns = []
    for grouped_annotations, grouped_predictions in grouped_inputs.values():
        result_per_reference_id, _, _ = _iou_assignments_for_same_reference_id(
            grouped_annotations, grouped_predictions, iou_threshold
        )
        iou_assigns.append(result_per_reference_id)
    return np.concatenate(iou_assigns)


def get_true_false_positives_confidences(
    annotations: List[BoxOrPolygonAnnotation],
    predictions: List[BoxOrPolygonPrediction],
    iou_threshold: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """Matches annotations and predictions based on linear sum cost and returns the
    intersection-over-union values of the matched annotation-prediction pairs, subject
    to the specified IoU threshold. Note that annotations and predictions from
    different reference_ids will not be matched with one another.
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linear_sum_assignment.html

    Args:
        annotations: list of box or polygon annotations
        predictions: list of box or polygon predictions
        iou_threshold: the intersection-over-union threshold for an
            annotation-prediction pair to be considered a match.

    Returns:
        1D numpy array that contains the 1 if true positive and 0 if false positive
            for each prediction.
        1D numpy array of confidence values.
    """
    grouped_inputs = group_boxes_or_polygons_by_reference_id(
        annotations, predictions
    )
    true_false_positives = []
    confidences = []
    for grouped_annotations, grouped_predictions in grouped_inputs.values():
        _, _, pred_to_anno = _iou_assignments_for_same_reference_id(
            grouped_annotations, grouped_predictions, iou_threshold
        )
        true_false_positives.append(pred_to_anno > -1)
        confidences.extend([pred.confidence for pred in grouped_predictions])
    return np.concatenate(true_false_positives), np.array(confidences)


def num_true_positives(
    annotations: List[BoxOrPolygonAnnotation],
    predictions: List[BoxOrPolygonPrediction],
    iou_threshold: float,
) -> int:
    """Counts the number of annotations with a matching prediction.

    A prediction is considered a match for an annotation if it has not yet been
    matched to another annotation, its reference_id is the same as the
    annotation, and its IoU with the annotation is at least the iou_threshold.

    Args:
        annotations: list of box or polygon annotations
        predictions: list of box or polygon predictions
        iou_threshold: the intersection-over-union threshold for an
            annotation-prediction pair to be considered a match.

    Returns:
        The number of true positives (predictions that are matched to annotations).
    """
    iou_assigns = iou_assignments(annotations, predictions, iou_threshold)
    true_positives = len(iou_assigns)
    return true_positives


def label_match_wrapper(metric_fn):
    """Decorator to add the ability to only apply metric to annotations and
    predictions with matching labels.

    Args:
        metric_fn: Metric function that takes a list of annotations, a list
            of predictions, and optional args and kwargs.

    Returns:
        Metric function which can optionally enforce matching labels.
    """

    @wraps(metric_fn)
    def wrapper(
        annotations: List[BoxOrPolygonAnnotation],
        predictions: List[BoxOrPolygonPrediction],
        *args,
        enforce_label_match: bool = False,
        **kwargs,
    ) -> ScalarResult:
        # Simply return the metric if we are not enforcing label matches.
        if not enforce_label_match:
            return metric_fn(annotations, predictions, *args, **kwargs)

        # For each bin of annotations/predictions, compute the metric applied
        # only to that bin. Then aggregate results across all bins.
        grouped_inputs = group_boxes_or_polygons_by_label(
            annotations, predictions
        )
        metric_results = []
        for binned_annotations, binned_predictions in grouped_inputs.values():
            metric_result = metric_fn(
                binned_annotations, binned_predictions, *args, **kwargs
            )
            metric_results.append(metric_result)
        assert all(
            isinstance(r, ScalarResult) for r in metric_results
        ), "Expected every result to be a ScalarResult"
        return ScalarResult.aggregate(metric_results)

    return wrapper



import os
from urllib.parse import urljoin


def nucleus_url(sub_path: str):
    nucleus_base = os.environ.get(
        "NUCLEUS_DASHBOARD", "https://dashboard.scale.com/nucleus/"
    )
    extra_params = os.environ.get("NUCLEUS_DASH_PARAMS", "")
    return urljoin(nucleus_base, sub_path.lstrip("/") + extra_params)



from dataclasses import dataclass
from typing import Dict

from .constants import W_KEY, X_KEY, Y_KEY, Z_KEY


@dataclass
class Quaternion:
    """Quaternion objects are used to represent rotation.

    We use the Hamilton/right-handed quaternion convention, where
    ::

        i^2 = j^2 = k^2 = ijk = -1

    The quaternion represented by the tuple ``(x, y, z, w)`` is equal to
    ``w + x*i + y*j + z*k``.

    Parameters:
        x (float): The x value.
        y (float): The y value.
        x (float): The z value.
        w (float): The w value.
    """

    x: float
    y: float
    z: float
    w: float

    @classmethod
    def from_json(cls, payload: Dict[str, float]):
        """Instantiates quaternion object from schematized JSON dict payload."""
        return cls(
            payload[X_KEY], payload[Y_KEY], payload[Z_KEY], payload[W_KEY]
        )

    def to_payload(self) -> dict:
        """Serializes quaternion object to schematized JSON dict."""
        return {
            X_KEY: self.x,
            Y_KEY: self.y,
            Z_KEY: self.z,
            W_KEY: self.w,
        }



from dataclasses import dataclass, field
from typing import Dict, Optional

from ..connection import Connection
from .constants import ThresholdComparison


@dataclass
class ScenarioTestMetric:
    """A Scenario Test Metric is an evaluation function combined with a comparator and associated with a Scenario Test.
    Scenario Test Metrics serve as the basis when evaluating a Model on a Scenario Test.
    """

    scenario_test_id: str
    eval_function_id: str
    threshold: Optional[float]
    connection: Connection
    eval_func_arguments: Optional[Dict] = field(default_factory=dict)
    threshold_comparison: ThresholdComparison = (
        ThresholdComparison.GREATER_THAN_EQUAL_TO
    )

    def set_threshold(self, threshold: Optional[float] = None) -> None:
        """Sets the threshold of the metric to the new value passed in as a parameters.
        Attributes:
            threshold (str): The ID of the scenario test.
        """
        payload = {"threshold": threshold}
        response = self.connection.post(
            payload,
            f"validate/metric/set_threshold/{self.scenario_test_id}/{self.eval_function_id}",
        )
        self.threshold = response.get("threshold")



import click
from rich.console import Console
from rich.live import Live
from rich.table import Column, Table
from rich.tree import Tree

from cli.client import init_client
from cli.helpers.nucleus_url import nucleus_url
from cli.helpers.web_helper import launch_web_or_invoke
from nucleus import NucleusAPIError
from nucleus.validate import (
    AvailableEvalFunctions,
    ScenarioTestMetric,
    ThresholdComparison,
)


@click.group("tests", invoke_without_command=True)
@click.option("--web", is_flag=True, help="Launch browser")
@click.pass_context
def tests(ctx, web):
    """Scenario Tests allow you to test your Models

    https://dashboard.scale.com/nucleus/scenario-tests
    """
    launch_web_or_invoke("scenario-tests", ctx, web, list_tests)


@tests.command("list")
def list_tests():
    """List all your Scenario Tests"""
    console = Console()
    with console.status("Finding your Scenario Tests", spinner="dots4"):
        client = init_client()
        scenario_tests = client.validate.scenario_tests
        table = Table(
            Column("id", overflow="fold", min_width=24),
            "Name",
            "slice_id",
            Column("url", overflow="fold"),
            title=":chart_with_upwards_trend: Scenario Tests",
            title_justify="left",
        )
        for ut in scenario_tests:
            table.add_row(ut.id, ut.name, ut.slice_id, nucleus_url(ut.id))
    console.print(table)


def format_criterion(
    criterion: ScenarioTestMetric, eval_functions: AvailableEvalFunctions
):
    op_map = {
        ThresholdComparison.GREATER_THAN: ">",
        ThresholdComparison.GREATER_THAN_EQUAL_TO: ">=",
        ThresholdComparison.LESS_THAN: "<",
        ThresholdComparison.LESS_THAN_EQUAL_TO: "<=",
    }
    eval_function_name = eval_functions.from_id(
        criterion.eval_function_id
    ).name
    return f"{eval_function_name} {op_map[criterion.threshold_comparison]} {criterion.threshold}"
    pass


@tests.command("describe")
@click.argument("scenario-test-id", default=None, required=False)
@click.option(
    "--all", "-a", is_flag=True, help="View details about all Scenario Tests"
)
def describe_test(scenario_test_id, all):
    """View detailed information about a test or all tests"""
    console = Console()
    # scenario_test = client.validate.get_scenario_test(scenario_test_id)
    assert scenario_test_id or all, "Must pass a scenario_test_id or --all"
    client = init_client()
    scenario_tests = client.validate.scenario_tests
    if all:
        tree = Tree(":chart_with_upwards_trend: All Scenario Tests")
        with Live(
            "Fetching description of all Scenario Tests",
            vertical_overflow="visible",
        ) as live:
            for idx, ut in enumerate(scenario_tests):
                test_branch = tree.add(f"{idx}: Scenario Test")
                build_scenario_test_info_tree(client, ut, test_branch)
                live.update(tree)
    else:
        with console.status("Fetching Scenario Test information"):
            scenario_test = [
                ut for ut in scenario_tests if ut.id == scenario_test_id
            ][0]
            tree = Tree(":chart_with_upwards_trend: Scenario Test")
            build_scenario_test_info_tree(client, scenario_test, tree)
            console.print(tree)


def build_scenario_test_info_tree(client, scenario_test, tree):
    try:
        slc = client.get_slice(scenario_test.slice_id)
        info_branch = tree.add(":mag: Details")
        info_branch.add(f"id: '{scenario_test.id}'")
        info_branch.add(f"name: '{scenario_test.name}'")
        scenario_test_url = nucleus_url(scenario_test.id)
        info_branch.add(f"url: {scenario_test_url}")
        slice_url = nucleus_url(f"{slc.dataset_id}/{slc.slice_id}")
        slice_branch = tree.add(":cake: Slice")
        slice_branch.add(f"id: '{slc.id}'")
        slice_info = slc.info()
        slice_branch.add(f"name: '{slice_info['name']}'")
        slice_branch.add(f"len: {len(slc.items)}")
        slice_branch.add(f"url: {slice_url}")
        criteria = scenario_test.get_eval_functions()
        criteria_branch = tree.add(":crossed_flags: Criteria")
        for criterion in criteria:
            pretty_criterion = format_criterion(
                criterion, client.validate.eval_functions
            )
            criteria_branch.add(pretty_criterion)
    except NucleusAPIError as e:
        error_branch = tree.add(":x: Error")
        error_branch.add(f"detail: {str(e)}")



import urllib.request
from functools import wraps


def sanitize_field(field):
    return urllib.request.quote(field.encode("UTF-8"), safe="")


def sanitize_string_args(function):
    """Helper decorator that ensures that all arguments passed are url-safe."""

    @wraps(function)
    def sanitized_function(*args, **kwargs):
        sanitized_args = []
        sanitized_kwargs = {}
        for arg in args:
            if isinstance(arg, str):
                arg = sanitize_field(arg)
            sanitized_args.append(arg)
        for key, value in kwargs.items():
            if isinstance(value, str):
                value = sanitize_field(value)
            sanitized_kwargs[key] = value
        return function(*sanitized_args, **sanitized_kwargs)

    return sanitized_function



import json
import os.path
from collections import Counter
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Sequence

from .annotation import is_local_path
from .camera_params import CameraParams
from .constants import (
    BACKEND_REFERENCE_ID_KEY,
    CAMERA_PARAMS_KEY,
    IMAGE_URL_KEY,
    METADATA_KEY,
    ORIGINAL_IMAGE_URL_KEY,
    POINTCLOUD_URL_KEY,
    REFERENCE_ID_KEY,
    TYPE_KEY,
    UPLOAD_TO_SCALE_KEY,
    URL_KEY,
)


class DatasetItemType(Enum):
    IMAGE = "image"
    POINTCLOUD = "pointcloud"


@dataclass  # pylint: disable=R0902
class DatasetItem:  # pylint: disable=R0902
    """A dataset item is an image or pointcloud that has associated metadata.

    Note: for 3D data, please include a :class:`CameraParams` object under a key named
    "camera_params" within the metadata dictionary. This will allow for projecting
    3D annotations to any image within a scene.

    Args:
        image_location (Optional[str]): Required if pointcloud_location is not present:
          The location containing the image for the given row of data. This can be a local
          path, or a remote URL. Remote formats supported include any URL (``http://`` or
          ``https://``) or URIs for AWS S3, Azure, or GCS (i.e. ``s3://``, ``gcs://``).

        pointcloud_location (Optional[str]): Required if image_location is not present:
          The remote URL containing the pointcloud JSON. Remote formats supported include
          any URL (``http://`` or ``https://``) or URIs for AWS S3, Azure, or GCS (i.e.
          ``s3://``, ``gcs://``).

        reference_id (Optional[str]): A user-specified identifier to reference the
          item.

        metadata (Optional[dict]): Extra information about the particular
          dataset item. ints, floats, string values will be made searchable in
          the query bar by the key in this dict. For example, ``{"animal":
          "dog"}`` will become searchable via ``metadata.animal = "dog"``.

          Categorical data can be passed as a string and will be treated
          categorically by Nucleus if there are less than 250 unique values in the
          dataset. This means histograms of values in the "Insights" section and
          autocomplete within the query bar.

          Numerical metadata will generate histograms in the "Insights" section,
          allow for sorting the results of any query, and can be used with the
          modulo operator For example: metadata.frame_number % 5 = 0

          All other types of metadata will be visible from the dataset item detail
          view.

          It is important that string and numerical metadata fields are consistent
          - if a metadata field has a string value, then all metadata fields with
          the same key should also have string values, and vice versa for numerical
          metadata.  If conflicting types are found, Nucleus will return an error
          during upload!

          The recommended way of adding or updating existing metadata is to re-run
          the ingestion (dataset.append) with update=True, which will replace any
          existing metadata with whatever your new ingestion run uses. This will
          delete any metadata keys that are not present in the new ingestion run.
          We have a cache based on image_location that will skip the need for a
          re-upload of the images, so your second ingestion will be faster than
          your first.

          For 3D (sensor fusion) data, it is highly recommended to include
          camera intrinsics the metadata of your camera image items. Nucleus
          requires these intrinsics to create visualizations such as cuboid
          projections. Refer to our `guide to uploading 3D data
          <https://nucleus.scale.com/docs/uploading-3d-data>`_ for more
          info.

          Coordinate metadata may be provided to enable the Map Chart in the Nucleus Dataset charts page.
          These values can be specified as `{ "lat": 52.5, "lon": 13.3, ... }`.

          Context Attachments may be provided to display the attachments side by side with the dataset
          item in the Detail View by specifying
          `{ "context_attachments": [ { "attachment": 'https://example.com/1' }, { "attachment": 'https://example.com/2' }, ... ] }`.

          .. todo ::
              Shorten this once we have a guide migrated for metadata, or maybe link
              from other places to here.

        upload_to_scale (Optional[bool]): Set this to false in order to use
          `privacy mode <https://nucleus.scale.com/docs/privacy-mode>`_.

          Setting this to false means the actual data within the item will not be
          uploaded to scale meaning that you can send in links that are only accessible
          to certain users, and not to Scale. Skipping upload to Scale is currently only
          implemented for images.
    """

    image_location: Optional[str] = None
    reference_id: str = (
        "DUMMY_VALUE"  # preserve argument ordering for backwards compatibility
    )
    metadata: Optional[dict] = None
    pointcloud_location: Optional[str] = None
    upload_to_scale: Optional[bool] = True

    def __post_init__(self):
        assert self.reference_id != "DUMMY_VALUE", "reference_id is required."
        assert bool(self.image_location) != bool(
            self.pointcloud_location
        ), "Must specify exactly one of the image_location or pointcloud_location parameters"
        if (self.pointcloud_location) and not self.upload_to_scale:
            raise NotImplementedError(
                "Skipping upload to Scale is not currently implemented for pointclouds."
            )
        self.local = (
            is_local_path(self.image_location) if self.image_location else None
        )
        self.type = (
            DatasetItemType.IMAGE
            if self.image_location
            else DatasetItemType.POINTCLOUD
        )
        camera_params = (
            self.metadata.get(CAMERA_PARAMS_KEY, None)
            if self.metadata
            else None
        )
        self.camera_params = (
            CameraParams.from_json(camera_params) if camera_params else None
        )

    @classmethod
    def from_json(cls, payload: dict):
        """Instantiates dataset item object from schematized JSON dict payload."""
        image_url = payload.get(IMAGE_URL_KEY, None) or payload.get(
            ORIGINAL_IMAGE_URL_KEY, None
        )
        pointcloud_url = payload.get(POINTCLOUD_URL_KEY, None)

        # handle case when re-converting Scene.from_json
        url = payload.get(URL_KEY, None)
        if url and not image_url and not pointcloud_url:
            if url.split(".")[-1] in ("jpg", "png"):
                image_url = url
            elif url.split(".")[-1] in ("json",):
                pointcloud_url = url

        if BACKEND_REFERENCE_ID_KEY in payload:
            payload[REFERENCE_ID_KEY] = payload[BACKEND_REFERENCE_ID_KEY]
        return cls(
            image_location=image_url,
            pointcloud_location=pointcloud_url,
            reference_id=payload.get(REFERENCE_ID_KEY, None),
            metadata=payload.get(METADATA_KEY, {}),
            upload_to_scale=payload.get(UPLOAD_TO_SCALE_KEY, True),
        )

    def local_file_exists(self):
        # TODO: make private
        return os.path.isfile(self.image_location)

    def to_payload(self, is_scene=False) -> dict:
        """Serializes dataset item object to schematized JSON dict."""
        payload: Dict[str, Any] = {
            METADATA_KEY: self.metadata or {},
        }

        payload[REFERENCE_ID_KEY] = self.reference_id

        if is_scene:
            if self.image_location:
                payload[URL_KEY] = self.image_location
            elif self.pointcloud_location:
                payload[URL_KEY] = self.pointcloud_location
            payload[TYPE_KEY] = self.type.value
            if self.camera_params:
                payload[CAMERA_PARAMS_KEY] = self.camera_params.to_payload()
        else:
            assert (
                self.image_location
            ), "Must specify image_location for DatasetItems not in a LidarScene or VideoScene"
            payload[IMAGE_URL_KEY] = self.image_location
            payload[UPLOAD_TO_SCALE_KEY] = self.upload_to_scale

        return payload

    def to_json(self) -> str:
        """Serializes dataset item object to schematized JSON string."""
        return json.dumps(self.to_payload(), allow_nan=False)


def check_all_paths_remote(dataset_items: Sequence[DatasetItem]):
    for item in dataset_items:
        if item.image_location and is_local_path(item.image_location):
            raise ValueError(
                f"All paths must be remote, but {item.image_location} is either "
                "local, or a remote URL type that is not supported."
            )


def check_for_duplicate_reference_ids(dataset_items: Sequence[DatasetItem]):
    ref_ids = []
    for dataset_item in dataset_items:
        if dataset_item.reference_id is None:
            raise ValueError(
                f"Reference ID cannot be None. Encountered DatasetItem with no reference ID:\n{dataset_item}"
            )
        ref_ids.append(dataset_item.reference_id)
    if len(ref_ids) != len(set(ref_ids)):
        duplicates = {
            f"{key}": f"Count: {value}"
            for key, value in Counter(ref_ids).items()
            if value > 1
        }
        raise ValueError(
            f"Duplicate reference IDs found among dataset_items: {duplicates}"
        )



import json
import os
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    List,
    Sequence,
    Tuple,
)

from nucleus.async_utils import (
    FileFormData,
    FileFormField,
    FormDataContextHandler,
    make_many_form_data_requests_concurrently,
)

from .constants import DATASET_ID_KEY, IMAGE_KEY, ITEMS_KEY, UPDATE_KEY
from .dataset_item import DatasetItem
from .errors import NotFoundError
from .payload_constructor import construct_append_payload
from .upload_response import UploadResponse

if TYPE_CHECKING:
    from . import NucleusClient


class DatasetItemUploader:
    def __init__(self, dataset_id: str, client: "NucleusClient"):  # noqa: F821
        self.dataset_id = dataset_id
        self._client = client

    def upload(
        self,
        dataset_items: List[DatasetItem],
        batch_size: int = 5000,
        update: bool = False,
        local_files_per_upload_request: int = 10,
    ) -> UploadResponse:
        """

        Args:
            dataset_items: Items to Upload
            batch_size: How many items to pool together for a single request for items
             without files to upload
            local_files_per_upload_request: How many items to pool together for a single
                request for items with files to upload
            update: Update records instead of overwriting

        Returns:

        """
        local_items = []
        remote_items = []
        if local_files_per_upload_request > 10:
            raise ValueError("local_files_per_upload_request should be <= 10")

        # Check local files exist before sending requests
        for item in dataset_items:
            if item.local:
                if not item.local_file_exists():
                    raise NotFoundError()
                local_items.append(item)
            else:
                remote_items.append(item)

        agg_response = UploadResponse(json={DATASET_ID_KEY: self.dataset_id})

        async_responses: List[Any] = []

        if local_items:
            async_responses.extend(
                self._process_append_requests_local(
                    self.dataset_id,
                    items=local_items,
                    update=update,
                    local_files_per_upload_request=local_files_per_upload_request,
                )
            )

        remote_batches = [
            remote_items[i : i + batch_size]
            for i in range(0, len(remote_items), batch_size)
        ]

        if remote_batches:
            tqdm_remote_batches = self._client.tqdm_bar(
                remote_batches, desc="Remote file batches"
            )
            for batch in tqdm_remote_batches:
                responses = self._process_append_requests(
                    dataset_id=self.dataset_id,
                    payload=construct_append_payload(batch, update),
                    update=update,
                    batch_size=batch_size,
                )
                async_responses.extend(responses)

        for response in async_responses:
            agg_response.update_response(response)

        return agg_response

    def _process_append_requests_local(
        self,
        dataset_id: str,
        items: Sequence[DatasetItem],
        update: bool,
        local_files_per_upload_request: int,
    ):
        # Batch into requests
        requests = []
        batch_size = local_files_per_upload_request
        for i in range(0, len(items), batch_size):
            batch = items[i : i + batch_size]
            request = FormDataContextHandler(
                self.get_form_data_and_file_pointers_fn(batch, update)
            )
            requests.append(request)

        progressbar = self._client.tqdm_bar(
            total=len(requests),
            desc=f"Uploading {len(items)} items in {len(requests)} batches",
        )

        return make_many_form_data_requests_concurrently(
            self._client,
            requests,
            f"dataset/{dataset_id}/append",
            progressbar=progressbar,
        )

    def _process_append_requests(
        self,
        dataset_id: str,
        payload: dict,
        update: bool,
        batch_size: int = 20,
    ):
        items = payload[ITEMS_KEY]
        payloads = [
            {ITEMS_KEY: items[i : i + batch_size], UPDATE_KEY: update}
            for i in range(0, len(items), batch_size)
        ]
        return [
            self._client.make_request(
                payload,
                f"dataset/{dataset_id}/append",
            )
            for payload in payloads
        ]

    def get_form_data_and_file_pointers_fn(
        self, items: Sequence[DatasetItem], update: bool
    ) -> Callable[..., Tuple[FileFormData, Sequence[BinaryIO]]]:
        """Defines a function to be called on each retry.

        File pointers are also returned so whoever calls this function can
        appropriately close the files. This is intended for use with a
        FormDataContextHandler in order to make form data requests.
        """

        def fn():

            # For some reason, our backend only accepts this reformatting of items when
            # doing local upload.
            # TODO: make it just accept the same exact format as a normal append request
            # i.e. the output of construct_append_payload(items, update)
            json_data = []
            for item in items:
                item_payload = item.to_payload()
                item_payload[UPDATE_KEY] = update
                json_data.append(item_payload)

            form_data = [
                FileFormField(
                    name=ITEMS_KEY,
                    filename=None,
                    value=json.dumps(json_data, allow_nan=False),
                    content_type="application/json",
                )
            ]

            file_pointers = []
            for item in items:
                # I don't know of a way to use with, since all files in the request
                # need to be opened at the same time.
                # pylint: disable=consider-using-with
                image_fp = open(item.image_location, "rb")
                # pylint: enable=consider-using-with
                img_type = f"image/{os.path.splitext(item.image_location)[1].strip('.')}"
                form_data.append(
                    FileFormField(
                        name=IMAGE_KEY,
                        filename=item.image_location,
                        value=image_fp,
                        content_type=img_type,
                    )
                )
                file_pointers.append(image_fp)
            return form_data, file_pointers

        return fn



import asyncio
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, BinaryIO, Callable, Sequence, Tuple

import aiohttp
import nest_asyncio
from tqdm import tqdm

from nucleus.constants import DEFAULT_NETWORK_TIMEOUT_SEC
from nucleus.errors import NucleusAPIError
from nucleus.retry_strategy import RetryStrategy

from .logger import logger

if TYPE_CHECKING:
    from . import NucleusClient


@dataclass
class FileFormField:
    name: str
    filename: str
    value: BinaryIO
    content_type: str


FileFormData = Sequence[FileFormField]

UPLOAD_SEMAPHORE = asyncio.Semaphore(10)


class FormDataContextHandler:
    """A context handler for file form data that handles closing all files in a request.

    Why do I need to wrap my requests in such a funny way?

    1. Form data must be regenerated on each request to avoid errors
        see https://github.com/Rapptz/discord.py/issues/6531
    2. Files must be properly open/closed for each request.
    3. We need to be able to do 1/2 above multiple times so that we can implement retries
        properly.

    Write a function that returns a tuple of form data and file pointers, then pass it to the
    constructor of this class, and this class will handle the rest for you.
    """

    def __init__(
        self,
        form_data_and_file_pointers_fn: Callable[
            ..., Tuple[FileFormData, Sequence[BinaryIO]]
        ],
    ):
        self._form_data_and_file_pointer_fn = form_data_and_file_pointers_fn
        self._file_pointers = None

    def __enter__(self):
        (
            file_form_data,
            self._file_pointers,
        ) = self._form_data_and_file_pointer_fn()
        form = aiohttp.FormData()
        for field in file_form_data:
            form.add_field(
                name=field.name,
                filename=field.filename,
                value=field.value,
                content_type=field.content_type,
            )
        return form

    def __exit__(self, exc_type, exc_val, exc_tb):
        for file_pointer in self._file_pointers:
            file_pointer.close()


def get_event_loop():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:  # no event loop running:
        loop = asyncio.new_event_loop()
    else:
        nest_asyncio.apply(loop)
    return loop


def make_many_form_data_requests_concurrently(
    client: "NucleusClient",
    requests: Sequence[FormDataContextHandler],
    route: str,
    progressbar: tqdm,
):
    """
    Makes an async post request with form data to a Nucleus endpoint.

    Args:
        client: The client to use for the request.
        requests: Each requst should be a FormDataContextHandler object which will
            handle generating form data, and opening/closing files for each request.
        route: route for the request.
        progressbar: A tqdm progress bar to use for showing progress to the user.
    """
    loop = get_event_loop()
    return loop.run_until_complete(
        form_data_request_helper(client, requests, route, progressbar)
    )


async def form_data_request_helper(
    client: "NucleusClient",
    requests: Sequence[FormDataContextHandler],
    route: str,
    progressbar: tqdm,
):
    """
    Makes an async post request with files to a Nucleus endpoint.

    Args:
        client: The client to use for the request.
        requests: Each request should be a FormDataContextHandler object which will
            handle generating form data, and opening/closing files for each request.
        route: route for the request.
    """
    async with aiohttp.ClientSession() as session:
        tasks = [
            asyncio.ensure_future(
                _post_form_data(
                    client=client,
                    request=request,
                    route=route,
                    session=session,
                    progressbar=progressbar,
                )
            )
            for request in requests
        ]
        return await asyncio.gather(*tasks)


async def _post_form_data(
    client: "NucleusClient",
    request: FormDataContextHandler,
    route: str,
    session: aiohttp.ClientSession,
    progressbar: tqdm,
):
    """
    Makes an async post request with files to a Nucleus endpoint.

    Args:
        client: The client to use for the request.
        request: The request to make (See FormDataContextHandler for more details.)
        route: route for the request.
        session: The session to use for the request.
    """
    endpoint = f"{client.endpoint}/{route}"
    logger.info("Posting to %s", endpoint)

    async with UPLOAD_SEMAPHORE:
        for sleep_time in RetryStrategy.sleep_times() + [-1]:
            with request as form:
                async with session.post(
                    endpoint,
                    data=form,
                    auth=aiohttp.BasicAuth(client.api_key, ""),
                    timeout=DEFAULT_NETWORK_TIMEOUT_SEC,
                ) as response:
                    logger.info(
                        "API request has response code %s", response.status
                    )

                    try:
                        data = await response.json()
                    except aiohttp.client_exceptions.ContentTypeError:
                        # In case of 404, the server returns text
                        data = await response.text()
                    if (
                        response.status in RetryStrategy.statuses
                        and sleep_time != -1
                    ):
                        time.sleep(sleep_time)
                        continue

                    if response.status == 503:
                        raise TimeoutError(
                            "The request to upload your max is timing out, please lower local_files_per_upload_request in your api call."
                        )

                    if not response.ok:
                        raise NucleusAPIError(
                            endpoint,
                            session.post,
                            aiohttp_response=(
                                response.status,
                                response.reason,
                                data,
                            ),
                        )
                    progressbar.update(1)
                    return data



import abc
from typing import Dict

import numpy as np


class SegmentationMaskLoader(abc.ABC):
    @abc.abstractmethod
    def fetch(self, url: str) -> np.ndarray:
        pass


class DummyLoader(SegmentationMaskLoader):
    def fetch(self, url: str) -> np.ndarray:
        raise NotImplementedError(
            "This dummy loader has to be replaced with an actual implementation of an image loader"
        )


class InMemoryLoader(SegmentationMaskLoader):
    """We use this loader in the tests, this allows us to serve images from memory instead of fetching
    from a filesystem.
    """

    def __init__(self, url_to_array: Dict[str, np.ndarray]):
        self.url_to_array = url_to_array

    def fetch(self, url: str):
        array = self.url_to_array[url]
        return array
